// This code is derived from code distributed as part of Google Perftools.
// The original is available in from google-perftools-2.0 in the file
// src/base/atomicops-internals-x86.h.  This file was retrieved Feb 06, 2012 by
// Robert Escriva.

/* Copyright (c) 2006, Google Inc.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 * ---
 * Original Author: Sanjay Ghemawat (don't blame him for bugs in this derivative)
 */

#ifndef e_atomic_h_
#define e_atomic_h_

#include <stdint.h>
#ifdef _MSC_VER
#define _WINSOCKAPI_
#include <Windows.h>
#endif

// This struct is not part of the public API of this module; clients may not
// use it.
// Features of this x86.  Values may not be correct before main() is run,
// but are set conservatively.
struct AtomicOps_x86CPUFeatureStruct {
  bool has_amd_lock_mb_bug; // Processor has AMD memory-barrier bug; do lfence
                            // after acquire compare-and-swap.
  bool has_sse2;            // Processor has SSE2.
  bool has_cmpxchg16b;      // Processor supports cmpxchg16b instruction.
};
extern struct AtomicOps_x86CPUFeatureStruct AtomicOps_Internalx86CPUFeatures;

#ifdef _MSC_VER
#define ATOMICOPS_COMPILER_BARRIER() MemoryBarrier()
#else
#define ATOMICOPS_COMPILER_BARRIER() __asm__ __volatile__("" : : : "memory")
#endif


namespace e
{
namespace atomic
{

inline void
memory_barrier()
{
#ifdef _MSC_VER
    MemoryBarrier();
#else
    __asm__ __volatile__("mfence" : : : "memory");
#endif
}

///////////////////////////////////// Store ////////////////////////////////////

inline void
store_32_nobarrier(volatile uint32_t* ptr, uint32_t value)
{
    *ptr = value;
}

inline void
store_32_acquire(volatile uint32_t* ptr, uint32_t value)
{
    *ptr = value;
    memory_barrier();
}

inline void
store_32_release(volatile uint32_t* ptr, uint32_t value)
{
    ATOMICOPS_COMPILER_BARRIER();
    *ptr = value; // An x86 store acts as a release barrier.
    // See comments in 64-bit version of Release_Store(), below.
}

inline void
store_64_nobarrier(volatile uint64_t* ptr, uint64_t value)
{
    *ptr = value;
}

inline void
store_64_acquire(volatile uint64_t* ptr, uint64_t value)
{
    *ptr = value;
    memory_barrier();
}

inline void
store_64_release(volatile uint64_t* ptr, uint64_t value)
{
    ATOMICOPS_COMPILER_BARRIER();

    *ptr = value; // An x86 store acts as a release barrier
    // for current AMD/Intel chips as of Jan 2008.
    // See also Acquire_Load(), below.

    // When new chips come out, check:
    //  IA-32 Intel Architecture Software Developer's Manual, Volume 3:
    //  System Programming Guide, Chatper 7: Multiple-processor management,
    //  Section 7.2, Memory Ordering.
    // Last seen at:
    //   http://developer.intel.com/design/pentium4/manuals/index_new.htm
    //
    // x86 stores/loads fail to act as barriers for a few instructions (clflush
    // maskmovdqu maskmovq movntdq movnti movntpd movntps movntq) but these are
    // not generated by the compiler, and are rare.  Users of these instructions
    // need to know about cache behaviour in any case since all of these involve
    // either flushing cache lines or non-temporal cache hints.
}

template <typename P>
inline void
store_ptr_nobarrier(P* volatile* ptr, P* value)
{
    *ptr = value;
}

template <typename P>
inline void
store_ptr_acquire(P* volatile* ptr, P* value)
{
    *ptr = value;
    memory_barrier();
}

template <typename P>
inline void
store_ptr_release(P* volatile* ptr, P* value)
{
    ATOMICOPS_COMPILER_BARRIER();

    *ptr = value; // An x86 store acts as a release barrier
    // for current AMD/Intel chips as of Jan 2008.
    // See also Acquire_Load(), below.
}

template <typename P>
inline void
store_ptr_fullbarrier(P* volatile* ptr, P* value)
{
    ATOMICOPS_COMPILER_BARRIER();
    *ptr = value; // An x86 store acts as a release barrier
    memory_barrier();
}

///////////////////////////////////// Load /////////////////////////////////////

inline uint32_t
load_32_nobarrier(volatile const uint32_t* ptr)
{
    return *ptr;
}

inline uint32_t
load_32_acquire(volatile const uint32_t* ptr)
{
    uint32_t value = *ptr; // An x86 load acts as a acquire barrier.
    // See comments in 64-bit version of Release_Store(), below.
    ATOMICOPS_COMPILER_BARRIER();
    return value;
}

inline uint32_t
load_32_release(volatile const uint32_t* ptr)
{
    memory_barrier();
    return *ptr;
}

inline uint64_t
load_64_nobarrier(volatile const uint64_t* ptr)
{
    return *ptr;
}

inline uint64_t
load_64_acquire(volatile const uint64_t* ptr)
{
    uint64_t value = *ptr; // An x86 load acts as a acquire barrier,
    // for current AMD/Intel chips as of Jan 2008.
    // See also Release_Store(), above.
    ATOMICOPS_COMPILER_BARRIER();
    return value;
}

inline uint64_t
load_64_release(volatile const uint64_t* ptr)
{
    memory_barrier();
    return *ptr;
}

template <typename P>
inline P*
load_ptr_nobarrier(P* volatile const* ptr)
{
    return *ptr;
}

template <typename P>
inline P*
load_ptr_acquire(P* volatile const* ptr)
{
    P* value = *ptr; // An x86 load acts as a acquire barrier,
    // for current AMD/Intel chips as of Jan 2008.
    // See also Release_Store(), above.
    ATOMICOPS_COMPILER_BARRIER();
    return value;
}

template <typename P>
inline P*
load_ptr_release(P* volatile const* ptr)
{
    memory_barrier();
    return *ptr;
}

/////////////////////////////// Compare and Swap ///////////////////////////////

inline uint32_t
compare_and_swap_32_nobarrier(volatile uint32_t* ptr, uint32_t old_value, uint32_t new_value)
{
#ifdef _MSC_VER
    return InterlockedCompareExchange(ptr, old_value, new_value);
#else
    uint32_t prev;
    __asm__ __volatile__("lock; cmpxchgl %1,%2"
                         : "=a" (prev)
                         : "q" (new_value), "m" (*ptr), "0" (old_value)
                         : "memory");
    return prev;
#endif
}

inline uint32_t
compare_and_swap_32_acquire(volatile uint32_t* ptr, uint32_t old_value, uint32_t new_value)
{
#ifdef _MSC_VER
    return InterlockedCompareExchangeAcquire(ptr, old_value, new_value);
#else
    uint32_t x = compare_and_swap_32_nobarrier(ptr, old_value, new_value);

    if (AtomicOps_Internalx86CPUFeatures.has_amd_lock_mb_bug)
    {
        __asm__ __volatile__("lfence" : : : "memory");
    }

    return x;
#endif
}

inline uint32_t
compare_and_swap_32_release(volatile uint32_t* ptr, uint32_t old_value, uint32_t new_value)
{
    return compare_and_swap_32_nobarrier(ptr, old_value, new_value);
}

inline uint64_t
compare_and_swap_64_nobarrier(volatile uint64_t* ptr, uint64_t old_value, uint64_t new_value)
{
#ifdef _MSC_VER
    return InterlockedCompareExchange64((volatile __int64*)ptr, old_value, new_value);
#else
    uint64_t prev;
    __asm__ __volatile__("lock; cmpxchgq %1,%2"
                         : "=a" (prev)
                         : "q" (new_value), "m" (*ptr), "0" (old_value)
                         : "memory");
    return prev;
#endif
}

inline uint64_t
compare_and_swap_64_acquire(volatile uint64_t* ptr, uint64_t old_value, uint64_t new_value)
{
#ifdef _MSC_VER
    return InterlockedCompareExchangeAcquire64((volatile __int64*)ptr, old_value, new_value);
#else
    uint64_t x = compare_and_swap_64_nobarrier(ptr, old_value, new_value);

    if (AtomicOps_Internalx86CPUFeatures.has_amd_lock_mb_bug)
    {
        __asm__ __volatile__("lfence" : : : "memory");
    }

    return x;
#endif
}

inline uint64_t
compare_and_swap_64_release(volatile uint64_t* ptr, uint64_t old_value, uint64_t new_value)
{
#ifdef _MSC_VER
    return InterlockedCompareExchangeRelease64((volatile __int64*)ptr, old_value, new_value);
#else
    return compare_and_swap_64_nobarrier(ptr, old_value, new_value);
#endif
}

template <typename P>
inline P*
compare_and_swap_ptr_nobarrier(P* volatile* ptr, P* old_value, P* new_value)
{
#ifdef _MSC_VER
    return (P*)InterlockedCompareExchangePointer((PVOID *)ptr, (PVOID)old_value, (PVOID)new_value);
#else
    P* prev;
    __asm__ __volatile__("lock; cmpxchgq %1,%2"
                         : "=a" (prev)
                         : "q" (new_value), "m" (*ptr), "0" (old_value)
                         : "memory");
    return prev;
#endif
}

template <typename P>
inline P*
compare_and_swap_ptr_acquire(P* volatile* ptr, P* old_value, P* new_value)
{
#ifdef _MSC_VER
    return (P*)InterlockedCompareExchangePointerAcquire((PVOID *)ptr, (PVOID)old_value, (PVOID)new_value);
#else
    P* x = compare_and_swap_nobarrier(ptr, old_value, new_value);

    if (AtomicOps_Internalx86CPUFeatures.has_amd_lock_mb_bug)
    {
        __asm__ __volatile__("lfence" : : : "memory");
    }

    return x;
#endif
}

template <typename P>
inline P*
compare_and_swap_ptr_release(P* volatile* ptr, P* old_value, P* new_value)
{
#ifdef _MSC_VER
	return (P*)InterlockedCompareExchangePointerRelease((PVOID *)ptr, (PVOID)old_value, (PVOID)new_value);
#else
    return compare_and_swap_ptr_nobarrier(ptr, old_value, new_value);
#endif
}

template <typename P>
inline P*
compare_and_swap_ptr_fullbarrier(P* volatile* ptr, P* old_value, P* new_value)
{
    P* x = compare_and_swap_ptr_nobarrier(ptr, old_value, new_value);

    if (AtomicOps_Internalx86CPUFeatures.has_amd_lock_mb_bug)
    {
        __asm__ __volatile__("lfence" : : : "memory");
    }

    return x;
}

//////////////////////////////// Atomic Exchange ///////////////////////////////

inline uint32_t
exchange_32_nobarrier(volatile uint32_t* ptr, uint32_t new_value)
{
#ifdef _MSC_VER
    InterlockedExchange(ptr, new_value);
#else
    __asm__ __volatile__("xchgl %1,%0"  // The lock prefix is implicit for xchg.
                         : "=r" (new_value)
                         : "m" (*ptr), "0" (new_value)
                         : "memory");
#endif
    return new_value;  // Now it's the previous value.
}

inline uint64_t
exchange_64_nobarrier(volatile uint64_t* ptr, uint64_t new_value)
{
#ifdef _MSC_VER
    InterlockedExchange64((volatile __int64*)ptr, new_value);
#else
    __asm__ __volatile__("xchgq %1,%0"  // The lock prefix is implicit for xchg.
                         : "=r" (new_value)
                         : "m" (*ptr), "0" (new_value)
                         : "memory");
#endif
    return new_value;  // Now it's the previous value.
}

template <typename P>
inline P*
exchange_ptr_nobarrier(P* volatile* ptr, P* new_value)
{
#ifdef _MSC_VER
    InterlockedExchangePointer(ptr, new_value);
#else
    __asm__ __volatile__("xchgq %1,%0"  // The lock prefix is implicit for xchg.
                         : "=r" (new_value)
                         : "m" (*ptr), "0" (new_value)
                         : "memory");
#endif
    return new_value;  // Now it's the previous value.
}

/////////////////////////////// Atomic Increment ///////////////////////////////

inline uint32_t
increment_32_nobarrier(volatile uint32_t* ptr, uint32_t increment)
{
#ifdef _MSC_VER
	return InterlockedExchangeAdd((LONG *)ptr, increment);
#else
    uint32_t temp = increment;
    __asm__ __volatile__("lock; xaddl %0,%1"
                         : "+r" (temp), "+m" (*ptr)
                         : : "memory");
    // temp now holds the old value of *ptr
    return temp + increment;
#endif
}

inline uint32_t
increment_32_fullbarrier(volatile uint32_t* ptr, uint32_t increment)
{
#ifdef _MSC_VER
    return InterlockedExchangeAdd((LONG *)ptr, increment);
#else
    uint32_t temp = increment;
    __asm__ __volatile__("lock; xaddl %0,%1"
                         : "+r" (temp), "+m" (*ptr)
                         : : "memory");

    // temp now holds the old value of *ptr
    if (AtomicOps_Internalx86CPUFeatures.has_amd_lock_mb_bug)
    {
        __asm__ __volatile__("lfence" : : : "memory");
    }

    return temp + increment;
#endif
}

inline uint64_t
increment_64_nobarrier(volatile uint64_t* ptr, uint64_t increment)
{
#ifdef _MSC_VER
    return InterlockedExchangeAdd64((LONG64 *) ptr, increment);
#else
    uint64_t temp = increment;
    __asm__ __volatile__("lock; xaddq %0,%1"
                         : "+r" (temp), "+m" (*ptr)
                         : : "memory");
    // temp now contains the previous value of *ptr
    return temp + increment;
#endif
}

inline uint64_t
increment_64_fullbarrier(volatile uint64_t* ptr, uint64_t increment)
{
#ifdef _MSC_VER
    return InterlockedExchangeAdd64((LONG64 *)ptr, increment);
#else
    uint64_t temp = increment;
    __asm__ __volatile__("lock; xaddq %0,%1"
                         : "+r" (temp), "+m" (*ptr)
                         : : "memory");

    // temp now contains the previous value of *ptr
    if (AtomicOps_Internalx86CPUFeatures.has_amd_lock_mb_bug)
    {
        __asm__ __volatile__("lfence" : : : "memory");
    }

    return temp + increment;
#endif
}

/////////////////////////////////// Atomic Or //////////////////////////////////

inline void
or_32_nobarrier(volatile uint32_t* ptr, uint32_t orwith)
{
#ifdef _MSC_VER
	_InterlockedOr((LONG *)ptr, (LONG)orwith);
#else
    __asm__ __volatile__("lock; orl %0,%1"
                         : "+r" (orwith), "+m" (*ptr)
                         : : "memory");
#endif
}

inline void
or_64_nobarrier(volatile uint64_t* ptr, uint64_t orwith)
{
#ifdef _MSC_VER
	InterlockedOr64((__int64*)ptr, orwith);
#else
    __asm__ __volatile__("lock; orq %0,%1"
                         : "+r" (orwith), "+m" (*ptr)
                         : : "memory");
#endif
}

////////////////////////////////// Atomic And //////////////////////////////////

inline void
and_32_nobarrier(volatile uint32_t* ptr, uint32_t andwith)
{
#ifdef _MSC_VER
	_InterlockedAnd((LONG *)ptr, (LONG)andwith);
#else
    __asm__ __volatile__("lock; andl %0,%1"
                         : "+r" (andwith), "+m" (*ptr)
                         : : "memory");
#endif
}

inline void
and_64_nobarrier(volatile uint64_t* ptr, uint64_t andwith)
{
#ifdef _MSC_VER
	InterlockedAnd64((volatile __int64*)ptr, andwith);
#else
    __asm__ __volatile__("lock; andq %0,%1"
                         : "+r" (andwith), "+m" (*ptr)
                         : : "memory");
#endif
}

} // namespace atomic
} // namespace e

#undef ATOMICOPS_COMPILER_BARRIER

#endif // e_atomic_h_
